package edu.kit.ipd.descartes.adaptation;

import java.io.IOException;

import org.apache.log4j.Logger;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;

import edu.kit.ipd.descartes.adaptation.exceptions.OperationNotPerformedException;
import edu.kit.ipd.descartes.adaptation.util.OclEvaluationHelper;
import edu.kit.ipd.descartes.mm.adaptation.AbstractControlFlowElement;
import edu.kit.ipd.descartes.mm.adaptation.Action;
import edu.kit.ipd.descartes.mm.adaptation.ActionReference;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationPlan;
import edu.kit.ipd.descartes.mm.adaptation.BranchAction;
import edu.kit.ipd.descartes.mm.adaptation.LoopAction;
import edu.kit.ipd.descartes.mm.adaptation.StartAction;
import edu.kit.ipd.descartes.mm.adaptation.StopAction;
import edu.kit.ipd.descartes.mm.adaptation.WeightedTactic;
import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationPoint;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerrepositoryPackage;
import edu.kit.ipd.descartes.mm.resourcelandscape.ResourcelandscapePackage;

public class TacticExecutor {

	static Logger logger = Logger.getLogger(TacticExecutor.class);

	private IActionHandler dmmModelActionHandler;
	private TacticsHistory hist;

	public TacticExecutor(IActionHandler actionExecutor) {
		dmmModelActionHandler = actionExecutor;
		hist = TacticsHistory.getInstance();
	}
	
	/**
     * Executes the given {@link AdaptationPlan}.
     * 
     * @param plan
     *            The adaptation plan to execute
     * @throws IOException
     * @throws OperationNotPerformedException
     */
    private void executeAdaptationPlan(AdaptationPlan plan) throws OperationNotPerformedException {

        StartAction start = null;
        start = findStartAction(plan);
        if (start == null) 
            throw new OperationNotPerformedException("No start action for adaptation plan " + plan.getName() + " found.");

        logger.info("Executing the adaptation plan <<" + plan.getName() + ">>.");
        executeNextStep(start.getSuccessor());
        logger.info("Adaptation plan of adaptation plan <<" + plan.getName() + ">> terminated.");
    }

	private StartAction findStartAction(AdaptationPlan plan) {
		for (AbstractControlFlowElement abstractControlFlowElement : plan.getSteps()) {
			// Find start of adaptation plan
			if (abstractControlFlowElement instanceof StartAction) {
				return (StartAction) abstractControlFlowElement;
			}
		}
		return null;
	}

	private void executeNextStep(AbstractControlFlowElement step)
			throws OperationNotPerformedException {

		if (step instanceof StopAction) {
			return;
		} else

		if (step instanceof BranchAction) {
		    executeBranchAction(step);
		} else
		
		if (step instanceof LoopAction) {
		    execLoopAction(step);
		} else

		if (step instanceof ActionReference) {
			executeReferredAction(step);
		}
		// Handle invalid actions and control flow elements
		else
		    throw new OperationNotPerformedException(
		            "No valid control flow element " + step.toString());
		
		// in any case (except stop action)
		// continue with the next step 
		executeNextStep(step.getSuccessor());
	}

    private void executeBranchAction(AbstractControlFlowElement branchAction) throws OperationNotPerformedException {
        BranchAction branch = (BranchAction) branchAction;
        if (isTrue(branch.getContext(), branch.getCondition()))
            executeAdaptationPlan(branch.getConditionTrueBranch());
        else
            executeAdaptationPlan(branch.getConditionFalseBranch());
    }

    private boolean isTrue(EObject obj, String branchCondition) {

        EClass context = null;

        switch (obj.eClass().getClassifierID()) {
        case ResourcelandscapePackage.RUNTIME_ENVIRONMENT:
            context = ResourcelandscapePackage.Literals.RUNTIME_ENVIRONMENT;
            break;
        case ContainerrepositoryPackage.CONTAINER_TEMPLATE:
            context = ContainerrepositoryPackage.Literals.CONTAINER_TEMPLATE;
            break;
         // TODO add further contexts if necessary
        default:
            logger.error("Could not set context for eClass " + obj.getClass());
            return false;
        }

        boolean invariant = OclEvaluationHelper.evaluateOclConstraint(obj, branchCondition, context);
        return invariant;
    }

    private void executeReferredAction(AbstractControlFlowElement step) throws OperationNotPerformedException {
        ActionReference ref = (ActionReference) step;
        Action action = ref.getRefersTo();
        logger.info("Executing action <<" + action.getName() + ">>.");
        AdaptationPoint currentAdaptationPoint = action
        		.getReferredAdaptationPoint();

        dmmModelActionHandler.execute(currentAdaptationPoint,
        		action.getAdaptationActionOperation());
    }

    private void execLoopAction(AbstractControlFlowElement loopAction) throws OperationNotPerformedException {
        LoopAction loop = (LoopAction) loopAction;
        int numberOfIterations = loop.getCounter();
        int i = 1;

        while (i <= numberOfIterations) {
            logger.debug("Executing iteration " + i + " of loop action " + loop.getBody().getName());
            executeAdaptationPlan(loop.getBody());
            i++;
        }
    }

	/**
	 * Applies the the given tactic.
	 * 
	 * @param currentTactic
	 *            The tactic to execute
	 * @return
	 */
	public void applyTactic(WeightedTactic weightedTactic) {
		try {
            executeAdaptationPlanOfTactic(weightedTactic);
            logger.info("Tactic " + weightedTactic.getUsedTactic().getName()
                    + " successfully applied.");
            hist.add(weightedTactic, TacticsHistory.NO_RESULT, true);

            // if this point is reached everything was fine, so persist them (save the models)
            dmmModelActionHandler.persistActions();
            return;
        } catch (OperationNotPerformedException e) {
			logger.error("Tactic " + weightedTactic.getUsedTactic().getName()
					+ " could not be executed. Please check adaptation plan and/or log files for errors.", e);
			weightedTactic.setCurrentWeight(0.0);
			hist.add(weightedTactic, TacticsHistory.NO_RESULT, false);
            /*
             * TODO Something went wrong. 
             * 1) Detect error 
             * 2) reevaluate tactics
             * 3) apply different tactic.
             */
        }
	}

    public void executeAdaptationPlanOfTactic(WeightedTactic weightedTactic) throws OperationNotPerformedException {
        executeAdaptationPlan(weightedTactic.getUsedTactic().getImplementedPlan());
    }

}
